#include <Wire.h>
#include "ht16k33.h"
#include <Wire.h>
#include <EEPROM.h>
#include "RTClib.h"
#include <WiFi.h>
#include <WiFiClient.h>
#include <WiFiMulti.h>
#include <ezTime.h>
#include "asciifont-pinout12.h"
#include <HardwareSerial.h>
#include <DFMiniMp3.h>

//OTA Files
#include <ESPmDNS.h>
#include <WiFiUdp.h>
#include <ArduinoOTA.h>

//******************************************************
#define ACCESS_POINT_NAME "<YourSSID>"        
#define ACCESS_POINT_PASSWORD "YOURPASSWORD" 
#define HOSTNAME "TimeCircuits"
#define localTimeZone "Australia/Adelaide"  // Best place on earth...
#define otaPassword "outatime"
#define otaHostname "TimeCircuits"
#define mp3Volume 10
//******************************************************

#define timeCursorPin 39
#define amPin 35
#define pmPin 37

volatile unsigned long lastLedTime = 0;
bool secondLedOn = 0;
volatile unsigned long lastMp3PlayTime = 0;
bool soundPlayed = 0;

// Define the 2 display controllers
HT16K33 HT1;
HT16K33 HT2;

Timezone myTZ;
WiFiClient espClient;
WiFiMulti WiFiMulti;

//MP3 Module setup
HardwareSerial MP3Serial(1);
class Mp3Notify;
typedef DFMiniMp3<HardwareSerial, Mp3Notify> DfMp3;
DfMp3 dfmp3(MP3Serial);

char dateDay[3];
char dateMonth[4];
char dateYear[5];
char timeHour[3];
char timeMinute[3];
char timeSecond[3];

//*****************************************************
// MP3 Player Class
class Mp3Notify
{
public:
  static void PrintlnSourceAction(DfMp3_PlaySources source, const char* action)
  {
    if (source & DfMp3_PlaySources_Sd) 
    {
        Serial.print("SD Card, ");
    }
    if (source & DfMp3_PlaySources_Usb) 
    {
        Serial.print("USB Disk, ");
    }
    if (source & DfMp3_PlaySources_Flash) 
    {
        Serial.print("Flash, ");
    }
    Serial.println(action);
  }
  static void OnError([[maybe_unused]] DfMp3& mp3, uint16_t errorCode)
  {
    // see DfMp3_Error for code meaning
    Serial.println();
    Serial.print("Com Error ");
    Serial.println(errorCode);
  }
  static void OnPlayFinished([[maybe_unused]] DfMp3& mp3, [[maybe_unused]] DfMp3_PlaySources source, uint16_t track)
  {
    Serial.print("Play finished for #");
    Serial.println(track);  
  }
  static void OnPlaySourceOnline([[maybe_unused]] DfMp3& mp3, DfMp3_PlaySources source)
  {
    PrintlnSourceAction(source, "online");
  }
  static void OnPlaySourceInserted([[maybe_unused]] DfMp3& mp3, DfMp3_PlaySources source)
  {
    PrintlnSourceAction(source, "inserted");
  }
  static void OnPlaySourceRemoved([[maybe_unused]] DfMp3& mp3, DfMp3_PlaySources source)
  {
    PrintlnSourceAction(source, "removed");
  }
};
//*******************************************************************************

void ConfigureWifi(){
  int timeout = 0;
  Serial.println("Configuring Wifi");
  
  WiFi.mode(WIFI_STA);
  WiFi.hostname(HOSTNAME);
  
  // Not Wifi.begin because I had issues with newer ESP32 chips connecting - This works instead.
  WiFiMulti.addAP(ACCESS_POINT_NAME, ACCESS_POINT_PASSWORD);

  //while (WiFi.status() != WL_CONNECTED) { // As Above
  while(WiFiMulti.run() != WL_CONNECTED) {
    delay(500);
    Serial.print(".");
  }
  if (WiFi.status() == WL_CONNECTED) {
    Serial.println("");
    Serial.println("WiFi connected");
    Serial.print("IP address: ");
    Serial.println(WiFi.localIP());
  }
  updateNTP();
  if (timeStatus() == timeSet){
    Serial.println("Time set");
  }
}

void updateDay(){
  myTZ.dateTime("d").toCharArray(dateDay,3);
  myTZ.dateTime("M").toCharArray(dateMonth,4);
  myTZ.dateTime("Y").toCharArray(dateYear,5);

  HT1.set16Seg(0,dateMonth[0]-32);
  HT1.set16Seg(1,dateMonth[1]-32);
  HT1.set16Seg(2,dateMonth[2]-32);  

  HT1.set16Seg(3,dateDay[0]-32);
  HT1.set16Seg(4,dateDay[1]-32);
  
  HT2.set16Seg(0,dateYear[0]-32);
  HT2.set16Seg(1,dateYear[1]-32);
  HT2.set16Seg(2,dateYear[2]-32);
  HT2.set16Seg(3,dateYear[3]-32);
  
  HT1.sendLed();
  HT2.sendLed();
}

void updateTime(){
  myTZ.dateTime("h").toCharArray(timeHour,3);
  myTZ.dateTime("i").toCharArray(timeMinute,3);
  myTZ.dateTime("s").toCharArray(timeSecond,3);

  HT2.set16Seg(4,timeHour[0]-32);
  HT2.set16Seg(5,timeHour[1]-32);

  HT2.set16Seg(6,timeMinute[0]-32);
  HT2.set16Seg(7,timeMinute[1]-32);
  HT2.sendLed();
}

void setup(){
  Serial.begin(57600); // Serial Start
  EEPROM.begin(512);

  pinMode(timeCursorPin, OUTPUT);
  pinMode(amPin, OUTPUT);
  pinMode(pmPin, OUTPUT);

  delay(500);  // Add some delay or the MP3 module does not seem to play...
  dfmp3.begin();

  HT1.begin(0x00);
  HT1.define16segFont((uint16_t*) &fontTable);
 
  HT2.begin(0x01);
  HT2.define16segFont((uint16_t*) &fontTable);

  HT1.setBrightness(5);
  HT2.setBrightness(5);
 
  HT1.clearAll();
  HT2.clearAll();
 
  ConfigureWifi();

  if(!myTZ.setCache(400)) myTZ.setLocation(localTimeZone);
  
  dfmp3.setVolume(mp3Volume);
  delay(200);
  dfmp3.playMp3FolderTrack(1);

  
  delay(500);
  updateTime();
  updateDay();
  
  if (isAM(myTZ.now())){
    digitalWrite(amPin, HIGH);
    digitalWrite(pmPin, LOW);
  }else{
    digitalWrite(amPin, LOW);
    digitalWrite(pmPin, HIGH);
  }

  //Enable OTA Updates
  ArduinoOTA
    .onStart([]() {
      String type;
      if (ArduinoOTA.getCommand() == U_FLASH)
        type = "sketch";
      else // U_SPIFFS
        type = "filesystem";

      // NOTE: if updating SPIFFS this would be the place to unmount SPIFFS using SPIFFS.end()
      Serial.println("Start updating " + type);
    })
    .onEnd([]() {
      Serial.println("\nEnd");
    })
    .onProgress([](unsigned int progress, unsigned int total) {
      Serial.printf("Progress: %u%%\r", (progress / (total / 100)));
    })
    .onError([](ota_error_t error) {
      Serial.printf("Error[%u]: ", error);
      if (error == OTA_AUTH_ERROR) Serial.println("Auth Failed");
      else if (error == OTA_BEGIN_ERROR) Serial.println("Begin Failed");
      else if (error == OTA_CONNECT_ERROR) Serial.println("Connect Failed");
      else if (error == OTA_RECEIVE_ERROR) Serial.println("Receive Failed");
      else if (error == OTA_END_ERROR) Serial.println("End Failed");
    });

  ArduinoOTA.setPassword(otaPassword);
  ArduinoOTA.setHostname(otaHostname);
  ArduinoOTA.begin();  
}

void loop(){
  ArduinoOTA.handle();
  events(); // Process EZtime
  if (secondLedOn == 0 && millis()-lastLedTime > 500){
      secondLedOn = 1;
      digitalWrite(timeCursorPin, HIGH);
      lastLedTime = millis();
  }
  if (secondLedOn == 1 && millis()-lastLedTime > 500){
    secondLedOn = 0;
    digitalWrite(timeCursorPin, LOW);
    lastLedTime = millis();
  } 

  if (minuteChanged()){
    if (isAM(myTZ.now())){
      digitalWrite(amPin, HIGH);
      digitalWrite(pmPin, LOW);
    }else{
      digitalWrite(amPin, LOW);
      digitalWrite(pmPin, HIGH);
    }
    updateTime();
    updateDay();
  }
}
